//
//  MMExpression.m
//
//  Created by Herbert Ptzl on Wed Mar 06 2002.
//  Copyright (c) 2002 H.Ptzl. 
//
//  This file is part of MathMap.
//
//  MathMap is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  MathMap is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with MathMap; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

#import "MMExpression.h"
#import "MMInvocation.h"
#import "MMUserValue.h"

#import "tags.h"
#import "cgen.h"
#import "noise.h"

void
mathmap_get_pixel (mathmap_invocation_t *invocation, 
    int drawable_index, int frame, int x, int y, guchar *pixel)
{
    userval_t *userval;

    /* FIXME: assert that drawable_index is legal */

    userval = &invocation->uservals[drawable_index];
/*
    printf("[%p, %d,%d]>%d,%d,%p ", 
        invocation, x, y, 
        userval->v.image.width, userval->v.image.height, 
        userval->v.image.data);
*/
    if (userval->v.image.data == 0
        || x < 0 || x >= userval->v.image.width
        || y < 0 || y >= userval->v.image.height)
    {
        int i;

        for (i = 0; i < 4; ++i)
            pixel[i] = invocation->edge_color[i];
    }
    else
    {
        guchar *data = (guchar*)userval->v.image.data + 4 * x + y * userval->v.image.row_stride;
        int i;

        for (i = 0; i < 4; ++i)
            pixel[i] = data[i];
    }
//    printf("<%d,%d,%d,%d> ", pixel[0], pixel[1], pixel[2], pixel[3]);
}


@implementation MMExpression

+ (void)initialize {
    static BOOL firstTime = YES;

    if (firstTime)
    {
	init_builtins();
	init_tags();
	init_macros();
	init_noise();
	init_compiler();

	firstTime = NO;
    }
}

+ (id)expressionWithString:(NSString *)aString
{
    return [[[self alloc] initWithString: aString] autorelease];
}

+ (NSDictionary *)verifyExpressionString:(NSString *)aString
{
    if (check_mathmap([aString cString]))
	return nil;		// nil heisst alles ok
    return [NSDictionary dictionaryWithObjectsAndKeys:
			     [NSString stringWithCString: error_string], @"message",
			 [NSNumber numberWithInt:0], @"start",
			 [NSNumber numberWithInt:0], @"end",
			 nil];
}

- (id)initWithString:(NSString *)aString
{
    self = [super init];
    if (self) {
        _expression = [aString retain];
        _mathmap = compile_mathmap([_expression cString],
            [[[NSBundle mainBundle] 
                pathForResource:@"new_template" 
                ofType:@"c"] cString]);
        assert(_mathmap != 0);     
    }
    return self;
}

- (id)initWithData:(NSData *)data encoding:(MMExpressionEncoding)encoding
{
    return nil;
}


- (void) dealloc
{
    [_expression release];
    free_mathmap(_mathmap);
    [super dealloc];
}


- (NSString *)string	{ return _expression; }

- (BOOL)isEqual:(id)object {
    if ([object isKindOfClass:[self class]]) {
        return [[self string] isEqual:[object string]];
    } else return NO;
}

- (NSData *)dataUsingEncoding:(MMExpressionEncoding)encoding
{
    return nil;
}

- (MMInvocation *)invocationWithUserValues:(NSArray *)values
    bitmapImageRep:(NSBitmapImageRep *)imageRep {
    MMInvocation *invocation;

    invocation = [[MMInvocation alloc] 
        initWithExpression:self
        userValues:values
        bitmapImageRep:imageRep];
    return [invocation autorelease];
}


- (NSArray *)defaultUserValues
{
    NSMutableArray *values = [NSMutableArray arrayWithCapacity: 16];
    userval_info_t *info;

    for (info = _mathmap->userval_infos; info != 0; info = info->next)
	[values addObject:[MMUserValue userValueWithCInfo:info]];
    return values;
}

- (NSArray *) userValuesWithProposedValues:(NSArray *)proposal {
    NSMutableArray *values = [NSMutableArray 
        arrayWithCapacity:[proposal count]];
    userval_info_t *info;

    for (info = _mathmap->userval_infos; info != 0; info = info->next) {
        MMUserValue *oldValue;
	MMUserValue *value = [MMUserValue userValueWithCInfo:info];
        NSEnumerator  *enumerator = [proposal objectEnumerator];
        
        while ((oldValue = [enumerator nextObject]) != nil) {
            if (([oldValue type] == [value type]) &&
                ([[oldValue name] isEqual:[value name]])) {
                switch ([value type]) {
                case MMIntUserValueType:
                case MMFloatUserValueType:
                    [value setFloatValue:[oldValue floatValue]];
                    break;
                    
                case MMBoolUserValueType:
                    [value setBoolValue:[oldValue boolValue]];
                    break;
                    
                case MMColorUserValueType:
                    [value setColor:[oldValue color]];
                    break;
                    
                case MMCurveUserValueType:
                case MMGradientUserValueType:
                    [value setCurve:[oldValue curve]];
                    break;
                    
                case MMImageUserValueType:
                    [value setImage:[oldValue image]];
                    break;
                }
                break;
            }
        }
        [values addObject:value];
    }
    return values;
}

@end

#import "MMTreeNode.h"
#import "UserDefaults.h"

@implementation MMExpression (PropertyList)

- (id)propertyList {
    NSDictionary *dict = [NSDictionary 
        dictionaryWithObjectsAndKeys:
            [[NSUserDefaults standardUserDefaults] 
                stringForKey:ExpressionLabelDefault], 
            EntryNameKey,
            ExpressionEntryKind, EntryKindKey,
            _expression, ValueKey,
            nil];
    return dict;
}

@end

#import "MMNodeData.h"

@implementation MMTreeNode (ExpressionHelper)

- (void)addExpressionToArray:(NSMutableArray *)expArray {
    MMNodeData *data = [self nodeData];
    
    if ([data type] == MMExpressionNodeType) {
        NSDictionary *dict = [data object];
        NSString *value = [dict objectForKey:ValueKey];
        [expArray addObject:value];
    }
}

@end
